/*
 * drivethres.cpp
 *
 * Home page of code is: http://smartmontools.sourceforge.net
 * Address of support mailing list: smartmontools-support@lists.sourceforge.net
 *
 * Copyright (C) 2003-11 Philip Williams, Bruce Allen
 * Copyright (C) 2008-12 Christian Franke <smartmontools-support@lists.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * You should have received a copy of the GNU General Public License
 * (for example COPYING); If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "config.h"
#include "int64.h"
#include <stdio.h>
#include "drivethres.h"
#include "utility.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef _WIN32
#include <io.h> // access()
#endif

#include <stdexcept>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>

#define MODEL_LEN 41
char attr_status;
const char * drivethres_cpp_cvsid = "$Id: drivethres.cpp 3460 2014-07-28 21:19:33Z chrfranke $"
                                     DRIVETHRES_H_CVSID;

//Configuration file for the threshold limits
static const char * configfile;

class drive_thres_cfg {
public:
  drive_thres_cfg(const char *);
 
  unsigned size() const
  { return m_thres_cfg_tab.size(); }

  const char * get_drive_name() 
  { return series_name; }

  const attr_thres_cfg & operator[](unsigned i)
  { return m_thres_cfg_tab[i]; } 

  void push_back(const attr_thres_cfg &);
private:
  char series_name[MODEL_LEN+1];
  std::vector<attr_thres_cfg> m_thres_cfg_tab;

};  

drive_thres_cfg::drive_thres_cfg(const char *vendor_name)
{
  strncpy(series_name, vendor_name, MODEL_LEN);
}


void drive_thres_cfg::push_back(const attr_thres_cfg & src_attr_cfg)
{
  attr_thres_cfg dest_attr_cfg = src_attr_cfg;
  m_thres_cfg_tab.push_back(dest_attr_cfg);
}

// Thershold data base 
static std::vector<drive_thres_cfg> thres_db;

// The function converts the arg to int and checks the range
// return val :
// -1 if the arg is null or value is not in the range
// value if the arg is valid one
static int GetValue(const char *arg,  const char *token, int lineno, const char *cfgfile,
		    int min, int max, int base)
{
  // make sure argument is there
  if (!arg) {
    pout("File %s line %d : Directive: %s takes integer argument from %d to %d.\n",
	     cfgfile, lineno, token, min, max);
    return -1;
  }

  // get argument value, check that passed token is in-range
  char *endptr;
  int val = strtol(arg,&endptr,base);

  if (!(!*endptr && min <= val && val <= max)) {
    pout("File %s line %d : Directive: %s has argument: %s; needs integer from %d to %d.\n",
	   cfgfile, lineno, token, arg, min, max);
    return -1;
  }

  // all is well; return value
  return val;
}

const char * const parseopts = "ITVF";
static const char * parseptr = parseopts;
// This function returns 1 if it has correctly parsed one token (and
// any arguments)returns 2 if the all tokens of the threshold are parsed
// for a attrtube id("-I -T -V -F"), else it returns -1 if an error was encountered and
// -2 if the error msg is already logged.
static int ParseToken(char * token, int lineno, attr_thres_cfg & attr_cfg)
{

  char sym;
  const char * delim = " \n\t";
  const char * arg = 0;

  // is the rest of the line a comment
  if (*token=='#')
    return 1;

  // is the token not recognized?
  if (*token!='-' || strlen(token)!=2)
    return -1;

  int val;
  // token we will be parsing:
  sym=token[1];
  // check the token is the expected one
  if (*parseptr!=sym) {
    pout("File %s line %d : ERROR Directive -%c Expected Directive -%c\n",
	     configfile, lineno, sym, *parseptr);
    return -2;
  }

  switch (sym) {
    case 'I':
      // Attribute ID whose thershold to be monitored
      if ((val = GetValue(arg=strtok(NULL,delim), token, lineno, configfile, 0, 255, 10)) < 0)
	return -2;
      attr_cfg.attr_id = val;
      parseptr++;
      break;
    case 'T':
      // Thershold values for the Attribute ID specified by -I
      if ((val = GetValue(arg=strtok(NULL,delim), token, lineno, configfile, 0, 0xFFFFFF, 16)) < 0)
	return -2;
      attr_cfg.attr_thres = val;
      parseptr++;
      break;
    case 'V':
      // Which value between raw and normalised value should be compared against
      // threshold specified by -T 
      attr_cfg.attr_flags = THRES_INVALID;
      arg = strtok(NULL, delim);
      if (!arg) {
	return -1;
      } else if (!strcmp("RAW", arg) ||
		 !strcmp("raw", arg)) {
	attr_cfg.attr_flags |= THRES_RAW;
      } else if (!strcmp("NORM", arg) ||
		 !strcmp("norm", arg)) {
	attr_cfg.attr_flags |= THRES_NORM;
      } else {
	return -1;
      }	
      parseptr++;
      break;
    case 'F':
      // Should the Error be reported when the value is above or below the threshold
      arg = strtok(NULL, delim);
      if (!arg) {
       return -1;	
      } else if (!strcmp("LOW", arg) ||
		 !strcmp("low", arg)) {
	attr_cfg.attr_flags |= THRES_LOW;
      } else if (!strcmp("HIGH", arg) ||
		 !strcmp("high", arg)) {
	attr_cfg.attr_flags |= THRES_HIGH;
      } else {
	return -1;
      }
      attr_cfg.attr_flags |= THRES_VALID;
      parseptr=parseopts;
      return 2;
      break;
    default:
      return -1;
  }
  return 1;
}

// This is the routine that adds things to the thres_db list.
//
// Return values are:
//  1: parsed a normal line and added a entry into thres_db.
//  0: found comment or blank line
// -1: found an error 

// Note: this routine modifies *line from the caller!
static int ParseConfigLine(int lineno, char * line)
{
  char *token=NULL;
  char *name=NULL;
  const char *delim = " \n\t";

  // get first token: device name. If a comment, skip line
  if (!(name=strtok(line,delim)) || *name=='#') {
    return 0;
  }

  // The entires for threshold database
  attr_thres_cfg  attr_cfg;
  drive_thres_cfg thres_cfg(name);

  // parse tokens one at a time from the file.
  while ((token=strtok(NULL,delim))){
    int retval=ParseToken(token, lineno, attr_cfg);

    if (retval>0) {
      // Parsed token
      if (retval==2)
        thres_cfg.push_back(attr_cfg);
      continue;
    }

    if (retval<0) {
      // error found on the line
      if (retval==-1)
        pout("File %s line %d : unknown Directive: %s\n."
             "Please refer to smartd_thres.conf for usage details.\n",
             configfile, lineno, token);
        return -1;
    }
  }
  thres_db.push_back(thres_cfg);
  return 1;
}

// Parses a configuration file.  Return values are:
//  N=>0: found N entries
// -1:    syntax error in config file
// -2:    config file does not exist
// -3:    config file exists but cannot be read
static int ParseThresConfigFile()
{
  // maximum line length in configuration file
  const int MAXLINELEN = 256;
  // maximum length of a continued line in configuration file
  const int MAXCONTLINE = 1023;

  int ret = -1;

  stdio_file f;

  if (!f.open(configfile,"r")) {
    // file exists but we can't read it 
    ret = (errno!=ENOENT ? -3 : -2);
    pout("%s: Unable to open configuration file %s\n",
	     strerror(errno),configfile);
    return ret;
  }

  // configuration file exists
  pout("Opened configuration file %s\n",configfile);

  // parse config file line by line
  int lineno = 1, cont = 0, contlineno = 0;
  char line[MAXLINELEN+2];
  char fullline[MAXCONTLINE+1];

  for (;;) {
    int len=0;
    char *lastslash;
    char *comment;
    contlineno++;

    // make debugging simpler
    memset(line,0,sizeof(line));

    // get a line, if we are at the end of line break
    if (!(fgets(line, MAXLINELEN+2, f))) 
      break;


    // See if line is too long
    len=strlen(line);
    if (len>MAXLINELEN){
      const char *warn;
      if (line[len-1]=='\n')
	warn="(including newline!) ";
      else
	warn="";
      pout("Error: line %d of file %s %sis more than MAXLINELEN=%d characters.\n",
	       (int)contlineno,configfile,warn,(int)MAXLINELEN);
      return -1;
    }

    // Ignore anything after comment symbol
    if ((comment=strchr(line,'#'))){
      *comment='\0';
      len=strlen(line);
    }

    // is the total line (made of all continuation lines) too long?
    if (cont+len>MAXCONTLINE){
      pout("Error: continued line %d (actual line %d) of file %s is more than MAXCONTLINE=%d characters.\n",
	       lineno, (int)contlineno, configfile, (int)MAXCONTLINE);
      return -1;
    }

    // copy string so far into fullline, and increment length
    strcpy(fullline+cont,line);
    cont+=len;

    // is this a continuation line.  If so, replace \ by space and look at next line
    if ( (lastslash=strrchr(line,'\\')) && !strtok(lastslash+1," \n\t")){
      *(fullline+(cont-len)+(lastslash-line))=' ';
      continue;
    }

    ret = ParseConfigLine( contlineno, fullline);
    if (ret<0) {
      // Error parsing the file close the file and clear the threshold db list
      f.close();	
      thres_db.clear();
      return -1;
    }
    
    lineno++;
    cont=0;

  }
  f.close();  
  return thres_db.size();
}

// Get the path of the threshold configuartion file
static const char * get_thresconfigfile_path()
{
#ifndef _WIN32
  return SMARTMONTOOLS_SYSCONFDIR"/smartd_thres.conf";
#else
  static std::string configfile_str = get_exe_dir() + "/smartd_thres.conf";
  return configfile_str.c_str();
#endif
}

int smartd_fd = -1;

static void smartd_fifo_init()
{

    int ret, sock_opt;
    const char *smartd_path = SMARTD_FIFO_PATH;

    /*
     * Create a named pipe for smartd data
     */
    umask(0);
    ret  = mkfifo(smartd_path, (S_IRWXU | S_IRWXG));
    if (ret > 0 && errno != EEXIST) {
	pout("unable to create smartd pipe %s"
		 " [errno = %d]", smartd_path, errno);
	goto smartd_error1;
    }

    smartd_fd = open(smartd_path, O_WRONLY|O_NONBLOCK);
    if (smartd_fd < 0) {
	pout("unable to open %s [errno = %d]",
		 smartd_path, errno);
	goto smartd_error1;
    }

    sock_opt = fcntl(smartd_fd, F_GETFL);
    if (sock_opt < 0) {
	pout("fcntl F_GETFL failed on %s [errno = %d]",
		 smartd_path, errno);
	goto smartd_error2;
    }

    sock_opt |= (O_NONBLOCK | O_TRUNC);
    ret = fcntl(smartd_fd, F_SETFL, sock_opt);
    if (ret < 0) {
	pout("fcntl F_SETFL failed on %s [errno = %d]",
		 smartd_path, errno);

	goto smartd_error2;
    }

    return;

smartd_error2:
    close(smartd_fd);
smartd_error1:
    smartd_fd = -1;
    return;
}


static void send_smartd_data(char status)
{
    int write_bytes;
    int ret_val;
    const char *vsmartd_pid_file = VSMARTD_PID_FILE;
    struct stat buffer;

    if (smartd_fd == -1) {
        smartd_fifo_init();
    }

    if (smartd_fd > 0) {
	if ((write_bytes = write(smartd_fd, &status, sizeof(char))) !=
	    sizeof(char)) {
	    pout("smartd data write error %s(%d)",
		    strerror(errno), errno);
	}
        /* 
         * lets now send SIGHUP to vsmartd, only if vsmartd is present
         */
        ret_val = stat("/var/run/vsmartd.pid",&buffer); /* check if present */
        if (ret_val == 0) {
            ret_val = pkill(vsmartd_pid_file, SIGHUP);
            if (ret_val <= 0) {
                pout(" Error:Could not tickle vsmartd  \n");
            }
        }
    }
}

static int get_pid_from_pid_file(const char  * name)
{
    int pid_file_fd;
    int ret_val;
    char pid_data[10];
    pid_file_fd = open(name, O_RDONLY);
    if (pid_file_fd < 0) {
        pout("unable to open %s [errno = %d]",
            name, errno);
        return -1;
    }
    ret_val = read(pid_file_fd, pid_data, 9);
    close(pid_file_fd);
    if (ret_val > 0) {
        /* we read something from the file, lets convert to int */
        pid_data[ret_val] = '\0';
        return atoi(pid_data);
    } else {
        return -1;
    }
}

/* sends signal to a process, after reading the pid from a file */
static int pkill(const char * name, int sig) 
{
    int p_pid;

    p_pid = get_pid_from_pid_file(name);
    if (p_pid > 0) {
        kill(p_pid, sig);
        return p_pid;
    } else {
        return -1;
    }
}



// This function parses the configuration file and builds the threshold database.
int ReadDriveThresConfig()
{
  //Empty the data base and  update new thresholds
  thres_db.clear(); 
  parseptr=parseopts;  
  configfile = get_thresconfigfile_path();

  return ParseThresConfigFile();
}

// check if the tresh_db is not empty
bool IsThresDbExists() 
{
  return !(thres_db.empty());
}

#define SMARTD_DEFAULT_DEVICE "DEFAULT_SSD"
// this function updates the threshold for the corresponding 
// attribube ID it returns true if successfull and false if
// if the attribute ID is not a valid one 
bool UpdateConfigThersholds(const char *device_name, 
			    const char * model, struct  ata_smart_values * smart_val,
			    struct ata_smart_threshold_cfg * attr_thres)
{
  // iterate over the thres_db list  
  for (unsigned i=0; i<thres_db.size(); i++) {
    // Get the config for one drive
    drive_thres_cfg thres_cfg=thres_db.at(i);
    // The series name to be configured
    const char * series_name=thres_cfg.get_drive_name();
    /* the config file may contain partial series name */
    if ((strstr(model, series_name)==NULL) &&
	(strcmp(SMARTD_DEFAULT_DEVICE, series_name)))
      continue;
    // iterate over the threshold config for the device and update the 
    // threshold values and flags
    for (unsigned j=0; j<thres_cfg.size(); j++) {
      // theshold config for each attribute configured
      attr_thres_cfg  attr_cfg=thres_cfg[j];
      unsigned char id=attr_cfg.attr_id;
      int idx=ata_find_attr_index(id, *smart_val);
      if (idx<0) {
        pout("Device %s:Error: The attribute ID: %d, is not valid for drive series %s\n",
		device_name, attr_cfg.attr_id, series_name); 
        return false;
      }
      attr_thres[idx].threshold=attr_cfg.attr_thres;
      attr_thres[idx].flags=attr_cfg.attr_flags;
    }
    return true;
  }
  return false;
}


void check_attr_cfg_thres(const char *drive_name,
			  const ata_smart_attribute & attr, int idx,
			  struct ata_smart_threshold_cfg * attr_thres,
			  const ata_vendor_attr_defs & defs) 
{
  uint64_t value=0;
  const char *delim = " \n\t";
  if (!(attr_thres[idx].flags & THRES_VALID))
    return;
  
  if (attr_thres[idx].flags & THRES_RAW)
    value = ata_get_attr_raw_value(attr, defs);
  else if (attr_thres[idx].flags & THRES_NORM)
    value = attr.current;

  if (((attr_thres[idx].flags & THRES_LOW) &&
      (value <= attr_thres[idx].threshold)) || 
      ((attr_thres[idx].flags & THRES_HIGH) &&
      (value >= attr_thres[idx].threshold))) {
    std::string attrname = ata_get_smart_attr_name(attr.id, defs);
    pout("Device %s:Failed SMART usage Attribute: %d %s.\n",drive_name,
	   attr.id, attrname.c_str());
    /* set the bit position correspoinding to the device */
    char *drive=strtok(drive_name, delim);
    attr_status|=1<<(drive[strlen(drive)-1]-'a');
  }
  return;
}

void set_attr_status()
{
    attr_status=0;
}

void send_attr_status()
{
    send_smartd_data(attr_status);
}
